/*
 * config_qos.c - QoS controller configuration handling
 *
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: maxim.menshikov@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifidefs.h>

#include "utils/debug.h"
#include "utils/utils.h"
#include "config.h"
#include "timer.h"
#if (EASYMESH_VERSION > 2)
#include "dpp.h"
#endif
#include "cntlr.h"
#include "config.h"
#include "easymesh.h"
#include "config_helpers.h"

#if (EASYMESH_VERSION > 2)

/* General API */
int clean_qos_rule_list(struct list_head *rule_list)
{
	struct qos_rule *p = NULL, *tmp;

	list_for_each_entry_safe(p, tmp, rule_list, list) {
		list_del(&p->list);
		free(p);
	}

	return 0;
}

int clean_qos_policy_list(struct list_head *rule_list)
{
	struct qos_mgmt_policy_elem *p = NULL, *tmp;

	list_for_each_entry_safe(p, tmp, rule_list, list) {
		list_del(&p->list);
		free(p);
	}

	return 0;
}

uint8_t cntlr_qos_diff(struct controller_config *cfg,
		struct controller_config *prev)
{
	uint8_t diff = 0;

	if (prev->qos.enabled != cfg->qos.enabled) {
		dbg("|%s:%d| qos enable state changed\n", __func__, __LINE__);
		diff |= CONFIG_DIFF_QOS;
	}

	return diff;
}

uint8_t cntlr_qos_rule_diff(struct controller_config *cfg,
		struct controller_config *prev)
{
	uint8_t diff = 0;

	if (list_memcmp(&prev->qos.rule, &cfg->qos.rule,
	                struct qos_rule,
	                sizeof(struct list_head))) {
		dbg("|%s:%d| qos rules have changed\n", __func__, __LINE__);
		diff |= CONFIG_DIFF_QOS;
	}

	return diff;
}

uint8_t cntlr_qos_mgmt_policy_diff(struct controller_config *cfg,
		struct controller_config *prev,
		bool mscs)
{
	uint8_t diff = 0;

	if (list_memcmp(mscs ? &prev->qos.policy.mscs_disallowed
	                     : &prev->qos.policy.scs_disallowed,
	                mscs ? &cfg->qos.policy.mscs_disallowed
	                     : &cfg->qos.policy.scs_disallowed,
	                struct qos_mgmt_policy_elem,
	                sizeof(struct list_head))) {
		dbg("|%s:%d| qos mgmt policies have changed\n", __func__, __LINE__);
		diff |= CONFIG_DIFF_QOS;
	}

	return diff;
}

/* QoS-related sections */
static int cntlr_config_get_qos_rule_dscp_pcp(struct controller_config *cc,
	struct uci_section *s,
	struct tlv_dscp_pcp *dscp_pcp)
{
	enum {
		QOS_RULE_DSCP_PCP_DSCP_PCP,
		NUM_QOS_RULE_DSCP_PCP_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_DSCP_PCP_DSCP_PCP] = {
			.name = "dscp_pcp",
			.type = UCI_TYPE_LIST
		},
	};
	struct uci_option *tb[NUM_QOS_RULE_DSCP_PCP_ATTRS];

	memset(dscp_pcp, 0, sizeof(struct tlv_dscp_pcp));

	uci_parse_section(s, opts, NUM_QOS_RULE_DSCP_PCP_ATTRS, tb);


	if (tb[QOS_RULE_DSCP_PCP_DSCP_PCP]) {
		struct uci_element *x;

		uci_foreach_element(&tb[QOS_RULE_DSCP_PCP_DSCP_PCP]->v.list, x) {
			int dscp_min = 0;
			int dscp_max = 0;
			int pcp = 0;

			if (sscanf(x->name, "%d-%d,%d", &dscp_min, &dscp_max, &pcp) == 3) {
				if (pcp <= 7 && pcp >= 0 && dscp_min >= 0 && dscp_max >= 0 &&
				    dscp_min < dscp_max &&
				    dscp_min < (sizeof(dscp_pcp->dscp_pcp) /
				                sizeof(dscp_pcp->dscp_pcp[0])) &&
				    dscp_max < (sizeof(dscp_pcp->dscp_pcp) /
				                sizeof(dscp_pcp->dscp_pcp[0]))) {
					uint8_t dscp;

					for (dscp = dscp_min; dscp <= dscp_max; ++dscp) {
						dscp_pcp->dscp_pcp[dscp] = pcp;
					}
				}
			} else if (sscanf(x->name, "%d,%d", &dscp_min, &pcp) == 2) {
				if (pcp <= 7 && pcp >= 0 && dscp_min >= 0 &&
				    dscp_min < (sizeof(dscp_pcp->dscp_pcp) /
				                sizeof(dscp_pcp->dscp_pcp[0]))) {
					dscp_pcp->dscp_pcp[dscp_min] = pcp;
				}
			}
		}
	}

	return 0;
}

static int cntlr_config_get_qos_rule_mscs(struct controller_config *cc,
                                          struct uci_section *s,
                                          struct mscs_desc_usr *mscs)
{
	enum {
		QOS_RULE_MSCS_UP_BITMAP,
		QOS_RULE_MSCS_UP_LIMIT,
		QOS_RULE_MSCS_STREAM_TIMEOUT,
		NUM_QOS_RULE_MSCS_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_MSCS_UP_BITMAP] = {
			.name = "up_bitmap",
			.type = UCI_TYPE_STRING },
		[QOS_RULE_MSCS_UP_LIMIT] = {
			.name = "up_limit",
			.type = UCI_TYPE_STRING },
		[QOS_RULE_MSCS_STREAM_TIMEOUT] = {
			.name = "stream_timeout",
			.type = UCI_TYPE_STRING
		},
	};
	struct uci_option *tb[NUM_QOS_RULE_MSCS_ATTRS];

	memset(mscs, 0, sizeof(struct mscs_desc_usr));

	uci_parse_section(s, opts, NUM_QOS_RULE_MSCS_ATTRS, tb);

#define TRANSFER_INT(__val_en, __field) \
	if (tb[(__val_en)]) {                               \
	    const char *val = tb[(__val_en)]->v.string;     \
	    mscs->__field = atoi(val);                      \
	}

	TRANSFER_INT(QOS_RULE_MSCS_UP_BITMAP, up_bitmap);
	TRANSFER_INT(QOS_RULE_MSCS_UP_LIMIT, up_limit);
	TRANSFER_INT(QOS_RULE_MSCS_STREAM_TIMEOUT, stream_timeout);

#undef TRANSFER_INT

	return 0;
}

static int cntlr_config_get_qos_rule_scs(struct controller_config *cc,
                                         struct uci_section *s,
                                         struct scs_desc_usr *scs)
{
	enum {
		QOS_RULE_SCS_ID,
		QOS_RULE_SCS_INTRA_ACCESS_PRIORITY,
		QOS_RULE_SCS_PROCESSING,
		NUM_QOS_RULE_SCS_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_SCS_ID] = {
			.name = "id",
			.type = UCI_TYPE_STRING
		},
		[QOS_RULE_SCS_INTRA_ACCESS_PRIORITY] = {
			.name = "intra_access_priority",
			.type = UCI_TYPE_STRING
		},
		[QOS_RULE_SCS_PROCESSING] = {
			.name = "processing",
			.type = UCI_TYPE_STRING
		},
	};
	struct uci_option *tb[NUM_QOS_RULE_SCS_ATTRS];

	memset(scs, 0, sizeof(struct scs_desc_usr));

	uci_parse_section(s, opts, NUM_QOS_RULE_SCS_ATTRS, tb);

	if (tb[QOS_RULE_SCS_ID]) {
		scs->scsid = atoi(tb[QOS_RULE_SCS_ID]->v.string);
	}

	if (tb[QOS_RULE_SCS_INTRA_ACCESS_PRIORITY]) {
		scs->intra_access_category_priority =
			atoi(tb[QOS_RULE_SCS_INTRA_ACCESS_PRIORITY]->v.string);
	}

	if (tb[QOS_RULE_SCS_PROCESSING]) {
		scs->processing = atoi(tb[QOS_RULE_SCS_PROCESSING]->v.string);
		scs->processing_present = true;
	}

	return 0;
}

static int cntlr_config_get_qos_rule_qm_dscp(struct controller_config *cc,
	struct uci_section *s,
	struct qos_mgmt_attr_usr *qm)
{
	enum {
		QOS_RULE_QM_DSCP_POLICY_ID,
		QOS_RULE_QM_DSCP_DSCP,
		NUM_QOS_RULE_QM_DSCP_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_QM_DSCP_POLICY_ID] = {
			.name = "policy_id",
			.type = UCI_TYPE_STRING },
		[QOS_RULE_QM_DSCP_DSCP] = {
			.name = "dscp",
			.type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_QOS_RULE_QM_DSCP_ATTRS];

	memset(&qm->dscp_policy, 0, sizeof(qm->dscp_policy));

	uci_parse_section(s, opts, NUM_QOS_RULE_QM_DSCP_ATTRS, tb);

	if (tb[QOS_RULE_QM_DSCP_POLICY_ID]) {
		const char *val = tb[QOS_RULE_QM_DSCP_POLICY_ID]->v.string;

		qm->dscp_policy.dscp_policy_id = atoi(val);
	}

	if (tb[QOS_RULE_QM_DSCP_DSCP]) {
		const char *val = tb[QOS_RULE_QM_DSCP_DSCP]->v.string;

		qm->dscp_policy.dscp = atoi(val);
	}

	return 0;
}

static int cntlr_config_get_qos_rule_qm(struct controller_config *cc,
                                        struct uci_section *s,
                                        struct qos_mgmt_attr_usr *qm)
{
	int ret = 0;
	enum {
		QOS_RULE_QM_TYPE,
		NUM_QOS_RULE_QM_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_QM_TYPE] = { .name = "policy_type", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_QOS_RULE_QM_ATTRS];

	memset(qm, 0, sizeof(struct qos_mgmt_attr_usr));

	uci_parse_section(s, opts, NUM_QOS_RULE_QM_ATTRS, tb);

	if (tb[QOS_RULE_QM_TYPE]) {
		const char *val = tb[QOS_RULE_QM_TYPE]->v.string;

		if (strcmp(val, "port_range") == 0)
			qm->attribute_id = QOS_MGMT_ATTR_ID_PORT_RANGE;
		else if (strcmp(val, "dscp_policy") == 0)
			qm->attribute_id = QOS_MGMT_ATTR_ID_DSCP_POLICY;
		else if (strcmp(val, "tclas") == 0)
			qm->attribute_id = QOS_MGMT_ATTR_ID_TCLAS;
		else if (strcmp(val, "domain_name") == 0)
			qm->attribute_id = QOS_MGMT_ATTR_ID_DOMAIN_NAME;
	}

	switch (qm->attribute_id)
	{
		case QOS_MGMT_ATTR_ID_DSCP_POLICY:
		{
			ret = cntlr_config_get_qos_rule_qm_dscp(cc, s, qm);
			break;
		}
		default:
		{
			ret = -1;
			break;
		}
	}

	return ret;
}

static int cntlr_config_get_qos_rule_mgmt(struct controller_config *cc,
                                          struct uci_section *s,
                                          struct qos_rule *rule)
{
	int ret = 0;
	enum {
		QOS_MGMT_RULE_QMID,
		QOS_MGMT_RULE_BSSID,
		QOS_MGMT_RULE_STA,
		NUM_QOS_MGMT_RULE_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_MGMT_RULE_QMID] = { .name = "qmid", .type = UCI_TYPE_STRING },
		[QOS_MGMT_RULE_BSSID] = { .name = "bssid", .type = UCI_TYPE_STRING },
		[QOS_MGMT_RULE_STA] = { .name = "sta", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_QOS_MGMT_RULE_ATTRS];

	uci_parse_section(s, opts, NUM_QOS_MGMT_RULE_ATTRS, tb);

	if (tb[QOS_MGMT_RULE_QMID]) {
		const char *val = tb[QOS_MGMT_RULE_QMID]->v.string;

		rule->mgmt.qmid = atoi(val);
	}

	if (tb[QOS_MGMT_RULE_BSSID]) {
		hwaddr_aton(tb[QOS_MGMT_RULE_BSSID]->v.string, rule->mgmt.bssid);
	}

	if (tb[QOS_MGMT_RULE_STA]) {
		hwaddr_aton(tb[QOS_MGMT_RULE_STA]->v.string, rule->mgmt.sta);
	}

	return ret;
}

static int cntlr_config_get_qos_rule(struct controller_config *cc,
                                     struct uci_section *s)
{
	int ret = 0;
	enum {
		QOS_RULE_ENABLED,
		QOS_RULE_TYPE,
		QOS_RULE_OUTPUT,
		QOS_RULE_ALWAYS_MATCH,
		NUM_QOS_RULE_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_RULE_ENABLED] = { .name = "enabled", .type = UCI_TYPE_STRING },
		[QOS_RULE_TYPE] = { .name = "type", .type = UCI_TYPE_STRING },
		[QOS_RULE_OUTPUT] = { .name = "output", .type = UCI_TYPE_STRING },
		[QOS_RULE_ALWAYS_MATCH] = { .name = "always_match",
		                            .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_QOS_RULE_ATTRS];
	struct qos_rule *rule;

	rule = calloc(1, sizeof(struct qos_rule));
	if (rule == NULL)
		return -1;

	uci_parse_section(s, opts, NUM_QOS_RULE_ATTRS, tb);

	if (tb[QOS_RULE_ENABLED]) {
		const char *val = tb[QOS_RULE_ENABLED]->v.string;

		rule->enabled = atoi(val) == 1 ? true : false;
	}

	if (tb[QOS_RULE_TYPE]) {
		const char *val = tb[QOS_RULE_TYPE]->v.string;

		if (!strcmp(val, "dscp_pcp"))
			rule->type = QOS_RULE_TYPE_DSCP_PCP;
		else if (!strcmp(val, "mscs"))
			rule->type = QOS_RULE_TYPE_MSCS;
		else if (!strcmp(val, "scs"))
			rule->type = QOS_RULE_TYPE_SCS;
		else if (!strcmp(val, "mgmt"))
			rule->type = QOS_RULE_TYPE_MGMT;
	}

	if (tb[QOS_RULE_OUTPUT]) {
		const char *val = tb[QOS_RULE_OUTPUT]->v.string;

		rule->output = atoi(val);
	}

	if (tb[QOS_RULE_ALWAYS_MATCH]) {
		const char *val = tb[QOS_RULE_ALWAYS_MATCH]->v.string;

		rule->always_match = atoi(val) ? true : false;
	}

	if (rule->type != QOS_RULE_TYPE_DSCP_PCP) {
		ret = cntlr_config_get_qos_rule_mgmt(cc, s, rule);
		if (ret != 0) {
			goto out;
		}
	}

	switch (rule->type)
	{
		case QOS_RULE_TYPE_DSCP_PCP:
		{
			ret = cntlr_config_get_qos_rule_dscp_pcp(cc, s, &rule->dscp_pcp);
			break;
		}
		case QOS_RULE_TYPE_MSCS:
		{
			ret = cntlr_config_get_qos_rule_mscs(cc, s, &rule->mgmt.mscs);
			break;
		}
		case QOS_RULE_TYPE_SCS:
		{
			ret = cntlr_config_get_qos_rule_scs(cc, s, &rule->mgmt.scs);
			break;
		}
		case QOS_RULE_TYPE_MGMT:
		{
			ret = cntlr_config_get_qos_rule_qm(cc, s, &rule->mgmt.qm);
			break;
		}
		default:
		{
			ret = -1;
			break;
		}
	}

out:
	if (ret == 0) {
		list_add_tail(&rule->list, &cc->qos.rule);
	} else {
		free(rule);
	}

	return ret;
}

static int cntlr_config_get_qos_policy(
	struct controller_config *cc,
	struct uci_section *s)
{
	enum {
		QOS_POLICY_MSCS_DISALLOWED,
		QOS_POLICY_SCS_DISALLOWED,
		NUM_QOS_POLICY_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_POLICY_MSCS_DISALLOWED] = { .name = "mscs_disallowed",
		                                 .type = UCI_TYPE_LIST },
		[QOS_POLICY_SCS_DISALLOWED] = { .name = "scs_disallowed",
		                                .type = UCI_TYPE_LIST },
	};
	struct uci_option *tb[NUM_QOS_POLICY_ATTRS];

	uci_parse_section(s, opts, NUM_QOS_POLICY_ATTRS, tb);

	INIT_LIST_HEAD(&cc->qos.policy.mscs_disallowed);
	INIT_LIST_HEAD(&cc->qos.policy.scs_disallowed);

	if (tb[QOS_POLICY_MSCS_DISALLOWED]) {
		struct uci_element *x;

		uci_foreach_element(&tb[QOS_POLICY_MSCS_DISALLOWED]->v.list, x) {
			struct qos_mgmt_policy_elem *el = calloc(1,
				sizeof(struct qos_mgmt_policy_elem));

			if (el == NULL ||
				hwaddr_aton(x->name, el->mac) == NULL) {
				free(el);
				return -1;
			}

			list_add(&el->list, &cc->qos.policy.mscs_disallowed);
		}
	}

	if (tb[QOS_POLICY_SCS_DISALLOWED]) {
		struct uci_element *x;

		uci_foreach_element(&tb[QOS_POLICY_SCS_DISALLOWED]->v.list, x) {
			struct qos_mgmt_policy_elem *el = calloc(1,
				sizeof(struct qos_mgmt_policy_elem));

			if (el == NULL ||
				hwaddr_aton(x->name, el->mac) == NULL) {
				free(el);
				return -1;
			}

			list_add(&el->list, &cc->qos.policy.scs_disallowed);
		}
	}

	return 0;
}

int cntlr_config_get_qos(struct controller_config *cc,
                         struct uci_section *s)
{
	int ret;
	enum {
		QOS_ENABLED,
		NUM_QOS_ATTRS
	};
	const struct uci_parse_option opts[] = {
		[QOS_ENABLED] = { .name = "enabled", .type = UCI_TYPE_STRING },
	};
	struct uci_option *tb[NUM_QOS_ATTRS];
	struct uci_element *e;

	uci_parse_section(s, opts, NUM_QOS_ATTRS, tb);

	if (tb[QOS_ENABLED]) {
		const char *val = tb[QOS_ENABLED]->v.string;

		cc->qos.enabled = atoi(val) == 1 ? true : false;
	}

	INIT_LIST_HEAD(&cc->qos.rule);

	uci_foreach_element(&s->package->sections, e) {
		struct uci_section *sect1 = uci_to_section(e);

		if (!strcmp(sect1->type, "qos_rule")) {
			ret = cntlr_config_get_qos_rule(cc, sect1);
		} else if (!strcmp(sect1->type, "qos_policy")) {
			ret = cntlr_config_get_qos_policy(cc, sect1);
		} else {
			ret = 0;
		}
		if (ret != 0)
			return ret;
	}

	return 0;
}

#endif
